#include "bic_container.h"
#include <string>
#include "properties.h"

extern void    errorCommit(const char * err);
extern void    messageCommit(const char * err);
using std::string;
BIC_Container::BIC_Container(void* buffer, double bufsize)
{
    this -> buffer = buffer;
    this -> bufsize = bufsize;

    void* ptr = (void*)&(((uint8_t*) buffer)[0]);
    if (((uint8_t*)ptr)[0] != BIC_VERSION_MAJOR)
    {
        string str = "BIC [v. " + std::to_string(((uint8_t*)ptr)[0]) + "." + std::to_string(((uint8_t*)ptr)[1]) +
                     "] doesn't equal runner [" + std::to_string(BIC_VERSION_MAJOR) + "." + std::to_string(BIC_VERSION_MINOR) + "]";
        errorCommit(str.c_str());
    }

    if (((uint8_t*)ptr)[1] != BIC_VERSION_MINOR)
    {
        string str = "BIC [v. " + std::to_string(((uint8_t*)ptr)[0]) + "." + std::to_string(((uint8_t*)ptr)[1]) +
                     "] doesn't equal runner [" + std::to_string(BIC_VERSION_MAJOR) + "." + std::to_string(BIC_VERSION_MINOR) + "]";
        messageCommit((string("BIC RUNNER [WARNING]: ") + str).c_str());
    }

    // Seek to our register header:
    ptr = (void*)&(((uint8_t*) buffer)[2]);
    regCount = ((uint8_t*) ptr)[0] + 1;
    ioCount = ((uint8_t*) ptr)[1];

    // Allocate space for our registers / IO ports
    registers = new port_t[regCount];
    for (uint8_t i = 0; i < regCount; ++i)
        registers[i] = port_t();

    io = new port_t[ioCount];
    for (uint8_t i = 0; i < ioCount; ++i)
        io[i] = port_t();

    seek = 4; // We are at byte 4
}

BIC_Container::~BIC_Container()
{
    if (registers != nullptr)
    {
        for (int i = 0; i < regCount; ++i)
            registers[i] = (void*) nullptr; // Delete all our pointers

        delete [] registers;
        registers = nullptr;
    }

    if (io != nullptr)
    {
        for (int i = 0; i < ioCount; ++i)
            io[i] = (void*) nullptr; // Delete all our pointers

        delete [] io;
        io = nullptr;
    }

    buffer = nullptr;
}

uint8_t BIC_Container::getBufU8()
{
        // A bit redundant here, but it should match other data type reading:
    return ((uint8_t*)&(((uint8_t*) buffer)[seek++]))[0];
}

uint16_t BIC_Container::getBufU16()
{
    uint16_t ret = ((uint16_t*)&(((uint8_t*) buffer)[seek]))[0];
    seek += 2;
    return ret;
}

int BIC_Container::getBufS32()
{
    int ret = ((int*)&(((uint8_t*) buffer)[seek]))[0];
    seek += 4;
    return ret;
}

double BIC_Container::getBufF64()
{
    double ret = ((double*)&(((uint8_t*) buffer)[seek]))[0];
    seek += 8;
    return ret;
}

string BIC_Container::getBufS()
{
    string str = "";
    while (seek < bufsize)
    {
        char val = ((char*) buffer)[seek++];
        if (val == 0)
            break;

        str += val;
    }
    return str;
}

void BIC_Container::regPrntTrace(uint8_t reg, port_t newPort)
{
    if (!tracing)
        return;

    for (auto it : regTraceList)
    {
         if (it == reg)
         {
            string str = "[BIC TRACE] Reg " + std::to_string(reg) + ": \n\t";
            port_t oldPort = getReg(reg);
            string oT, nT, oV, nV;
            if (oldPort.type == PT_float)
            {
                oT = "FLOAT to ";
                oV = std::to_string(*((double*) oldPort.getValue())) + " to ";
            }
            else if (oldPort.type == PT_int)
            {
                oT = "INT to ";
                oV = std::to_string(*((int*) oldPort.getValue())) + " to ";
            }
            else if (oldPort.type == PT_ptr)
            {
                oT = "PTR to ";
                oV = "[";
                for (int i = 0; i < std::min(5, (int) oldPort.length); ++i)
                {
                    if (oV == "[")
                        oV += std::to_string(((uint8_t*) oldPort.getValue())[i]);
                    else
                        oV += ", " + std::to_string(((uint8_t*) oldPort.getValue())[i]);
                }
                if (oldPort.length > 5)
                    oV += ", ...] to ";
                else
                    oV += "] to ";
            }
            else
            {
                oT = "UNDF to ";
                oV = "UNDF to ";
            }

            if (newPort.type == PT_float)
            {
                nT = "FLOAT\n\t";
                nV = std::to_string(*((double*) newPort.getValue()));
            }
            else if (newPort.type == PT_int)
            {
                nT = "INT\n\t";
                nV = std::to_string(*((int*) newPort.getValue()));
            }
            else if (newPort.type == PT_ptr)
            {
                nT = "PTR\n\t";
                nV = "[";
                for (int i = 0; i < std::min(5, (int) newPort.length); ++i)
                {
                    if (nV == "[")
                        nV += std::to_string(((uint8_t*) newPort.getValue())[i]);
                    else
                        nV += ", " + std::to_string(((uint8_t*) newPort.getValue())[i]);
                }
                if (newPort.length > 5)
                    nV += ", ...]";
                else
                    nV += "]";
            }
            else
            {
                nT = "UNDF\n\t";
                nV = "UNDF";
            }

            str += oT + nT;
            str += oV + nV;

            messageCommit(str.c_str());
            break;
         }
    }
}

void BIC_Container::ioPrntTrace(uint8_t io, port_t newPort)
{
    if (!tracing)
        return;

    for (auto it : ioTraceList)
    {
         if (it == io)
         {
            string str = "[BIC TRACE] IO " + std::to_string(io) + ": \n\t";
            port_t oldPort = getIO(io);
            string oT, nT, oV, nV;
            if (oldPort.type == PT_float)
            {
                oT = "FLOAT to ";
                oV = std::to_string(*((double*) oldPort.getValue())) + " to ";
            }
            else if (oldPort.type == PT_int)
            {
                oT = "INT to ";
                oV = std::to_string(*((int*) oldPort.getValue())) + " to ";
            }
            else if (oldPort.type == PT_ptr)
            {
                oT = "PTR to ";
                oV = "[";
                for (int i = 0; i < std::min(5, (int) oldPort.length); ++i)
                {
                    if (oV == "[")
                        oV += std::to_string(((uint8_t*) oldPort.getValue())[i]);
                    else
                        oV += ", " + std::to_string(((uint8_t*) oldPort.getValue())[i]);
                }
                if (oldPort.length > 5)
                    oV += ", ...] to ";
                else
                    oV += "] to ";
            }
            else
            {
                oT = "UNDF to ";
                oV = "UNDF to ";
            }

            if (newPort.type == PT_float)
            {
                nT = "FLOAT\n\t";
                nV = std::to_string(*((double*) newPort.getValue()));
            }
            else if (newPort.type == PT_int)
            {
                nT = "INT\n\t";
                nV = std::to_string(*((int*) newPort.getValue()));
            }
            else if (newPort.type == PT_ptr)
            {
                nT = "PTR\n\t";
                nV = "[";
                for (int i = 0; i < std::min(5, (int) newPort.length); ++i)
                {
                    if (nV == "[")
                        nV += std::to_string(((uint8_t*) newPort.getValue())[i]);
                    else
                        nV += ", " + std::to_string(((uint8_t*) newPort.getValue())[i]);
                }
                if (newPort.length > 5)
                    nV += ", ...]";
                else
                    nV += "]";
            }
            else
            {
                nT = "UNDF\n\t";
                nV = "UNDF";
            }

            str += oT + nT;
            str += oV + nV;

            messageCommit(str.c_str());
            break;
         }
    }
}

void BIC_Container::pullTraceReg(uint8_t id)
{
    uint32_t c = 0;
    for (auto it : regTraceList)
    {
        if (it == id)
        {
            regTraceList.erase(regTraceList.begin() + c);
            break;
        }

        ++c;
    }

    if (ioTraceList.size() <= 0 && regTraceList.size() <= 0)
        tracing = false;
}

void BIC_Container::pullTraceIO(uint8_t id)
{
    uint32_t c = 0;
    for (auto it : ioTraceList)
    {
        if (it == id)
        {
            regTraceList.erase(ioTraceList.begin() + c);
            break;
        }

        ++c;
    }

    if (ioTraceList.size() <= 0 && regTraceList.size() <= 0)
        tracing = false;
}
