#include "bic_runner.h"
#include <string>
#include <cmath>
#include <mutex>
#include "properties.h"
using std::string;
using std::to_string;

#ifdef BIC_INTERFACE_GAMEMAKER
extern void    (*GMAsyncTrigger)(int, int);
extern int     (*GMMapCreate)(int _num, ...);
extern bool    (*GMMapAddDouble)(int _index, const char* _pKey, double value);
extern bool    (*GMMapAddString) (int _index, const char* _pKey, const char* pVal);
extern const int GM_EVENT = 75;
extern void    errorCommit(const char * err);
extern void    messageCommit(const char * err);
extern std::mutex gmConnect;

void errorCommit(string err, vector <string> errs)
{
    string str = err;
    for (uint32_t i = 0; i < errs.size(); ++i)
    {
        str += "\n\t";
        str += errs[i];
    }
    errorCommit(str.c_str());
}
#endif // BIC_INTERFACE_GAMEMAKER

BIC_Runner::~BIC_Runner()
{
    if (running)
        stopThread();
}

void BIC_Runner::setWaitValue(port_t reg0)
{
    waitPort = reg0;
}

void BIC_Runner::pullTask(BIC_Container* task)
{
    for (uint32_t i = 0; i < tasks.size(); ++i)
    {
        if (tasks[i] == task)
        {
            tasks.erase(tasks.begin() + i);
            break;
        }
    }

}

void BIC_Runner::startThread()
{
    if (running)
    {
        errorCommit("Cannot start thread, thread already running!");
        return;
    }

    running = true;
    messageCommit("[BIC] Starting thread...");
    thread = new std::thread(BIC_Runner::threadExecute, this);

}

void BIC_Runner::taskCalculateBit(BIC_Container* task)
{
    // See if we are at the end of the buffer and reset it:
    if (task -> getSeek() >= task -> getBufSize())
    {
        // UNLIKE the internal runner, we pull tasks automatically
        string str = "BIC RUNNER [WARNING]: Unexpected end of buffer!\n\tseekpos: ";
        str += std::to_string(task -> getSeek());
        str += "\n\tbuffsize: ";
        str += std::to_string(task -> getBufSize());
        messageCommit(str.c_str());

        pullTask(task);
        task -> setSeek(4);

        // Send our call request to GameMaker
        #ifdef BIC_INTERFACE_GAMEMAKER
            gmConnect.lock();
            int mapID = GMMapCreate(0);
            GMMapAddDouble(mapID, "container_ptr", (double)((uint32_t) task));
            GMMapAddDouble(mapID, "runner_ptr", (double)((uint32_t) this));
            GMMapAddDouble(mapID, "function_call", task -> getCall());
            GMAsyncTrigger(mapID, GM_EVENT);
            gmConnect.unlock();
        #endif // BIC_INTERFACE_GAMEMAKER
        return;
    }

    // Grab our command ID:
    uint8_t command = task -> getBufU8();

    switch (command)
    {
        // end
        case 0:
        {
            pullTask(task);
            task -> setSeek(4);

            #ifdef BIC_INTERFACE_GAMEMAKER
            // Send our call request to GameMaker
            gmConnect.lock();
            int mapID = GMMapCreate(0);
            GMMapAddDouble(mapID, "container_ptr", (double)((uint32_t) task));
            GMMapAddDouble(mapID, "runner_ptr", (double)((uint32_t) this));
            GMMapAddDouble(mapID, "function_call", task -> getCall());
            GMAsyncTrigger(mapID, GM_EVENT);
            gmConnect.unlock();
            #endif // BIC_INTERFACE_GAMEMAKER
        }
        break;

        // srout
        case 1:
        {
            uint8_t ioN = task -> getBufU8(),
                    regN = task -> getBufU8();

            task -> setIO(ioN, (task -> getReg(regN)).mkCopy());
        }
        break;

        // sinr
        case 2:
        {
            uint8_t ioN = task -> getBufU8(),
                    regN = task -> getBufU8();

            task -> setReg(regN, (task -> getIO(ioN)).mkCopy());
        }
        break;

        // excall
        case 3:
        {
            uint16_t script = task -> getBufU16();

            // Send a request to GameMaker to call the script:
            setWaitValue(port_t());

            #ifdef BIC_INTERFACE_GAMEMAKER
            gmConnect.lock();
            int mapID = GMMapCreate(0);
            GMMapAddDouble(mapID, "external_call", script);
            GMMapAddDouble(mapID, "container_ptr", (double)((uint32_t) task));
            GMMapAddDouble(mapID, "runner_ptr", (double)((uint32_t) this));
            GMAsyncTrigger(mapID, GM_EVENT);
            gmConnect.unlock();
            #endif // BIC_INTERFACE_GAMEMAKER

            // Wait until we have our updated value sent back:
            while (waitPort.type == PT_undf) ;
            task -> setReg(0, waitPort.mkCopy());

        }
        break;

        //jumpp
        case 4:
        {
            uint16_t byte = task -> getBufU16();
            port_t port = task -> getReg(0);

            if (port.type > PT_int)
            {
                errorCommit("Expected a number!",
                            {"function: jumpp",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (port.type == PT_int && *((int*) port.getValue()) == 0)
                task -> setSeek(byte);

            if (port.type == PT_float && (long)(*((double*) port.getValue())) == 0)
                task -> setSeek(byte);
        }
        break;

        //jumpn
        case 5:
        {
            uint16_t byte = task -> getBufU16();
            port_t port = task -> getReg(0);

            if (port.type > PT_int)
            {
                errorCommit("Expected a number!",
                            {"function: jumpn",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (port.type == PT_int && *((int*) port.getValue()) != 0)
                task -> setSeek(byte);

            if (port.type == PT_float && (long)(*((double*) port.getValue())) != 0)
                task -> setSeek(byte);
        }
        break;

        //jumpnn
        case 6:
        {
            uint16_t byte = task -> getBufU16();
            port_t port = task -> getReg(0);

            if (port.type > PT_int)
            {
                errorCommit("Expected a number!",
                            {"function: jumpnn",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (port.type == PT_int && *((int*) port.getValue()) < 0)
                task -> setSeek(byte);

            if (port.type == PT_float && *((double*) port.getValue()) < 0)
                task -> setSeek(byte);
        }
        break;

        //jumppn
        case 7:
        {
            uint16_t byte = task -> getBufU16();
            port_t port = task -> getReg(0);

            if (port.type > PT_int)
            {
                errorCommit("Expected a number!",
                            {"function: jumppn",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (port.type == PT_int && *((int*) port.getValue()) > 0)
                task -> setSeek(byte);

            if (port.type == PT_float && *((double*) port.getValue()) > 0)
                task -> setSeek(byte);
        }
        break;

        //jump
        case 8:
        {
            uint16_t byte = task -> getBufU16();
            task -> setSeek(byte);
        }
        break;

        //jumpr
        case 9:
        {
            port_t port = task -> getReg(0);

            if (port.type != PT_int)
            {
                errorCommit("Expected type int!",
                            {"function: jumpr",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            task -> setSeek(*((int*) port.getValue()));
        }
        break;

        // rcpr
        case 10:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            task -> setReg(reg1N, (task -> getReg(reg2N)).mkCopy());
        }
        break;

        // rcpi
        case 11:
        {
            uint8_t reg1N = task -> getBufU8();
            int     value = task -> getBufS32();

            port_t port;
            port = new int (value);
            task -> setReg(reg1N, port);
        }
        break;

        // rcpf
        case 12:
        {
            uint8_t reg1N = task -> getBufU8();
            double  value = task -> getBufF64();

            port_t port;
            port = new double (value);
            task -> setReg(reg1N, port);
        }
        break;

        // rcps
        case 13:
        {
            uint8_t reg1N = task -> getBufU8();
            string  value = task -> getBufS();

            uint8_t* val = new uint8_t[value.length() + 1];
            for (uint32_t i = 0; i < value.length(); ++i)
                val[i] = value[i];

            val[value.length()] = 0;

            port_t port;
            port = val;
            port.length = value.length() + 1; // Count the null termination
            task -> setReg(reg1N, port);
        }
        break;

        // radd
        case 14:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != reg2.type ||
                reg1.type > PT_int)
            {
                errorCommit("Incorrect types!",
                            {"function: radd",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            port_t port;
            if (reg1.type == PT_float)
                port = new double(*((double*)reg1.getValue()) + *((double*)reg2.getValue()));

            if (reg1.type == PT_int)
                port = new int(*((int*)reg1.getValue()) + *((int*)reg2.getValue()));

            task -> setReg(0, port);
        }
        break;

        // rsub
        case 15:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != reg2.type ||
                reg1.type > PT_int)
            {
                errorCommit("Incorrect types!",
                            {"function: rsub",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            port_t port;
            if (reg1.type == PT_float)
                port = new double(*((double*)reg1.getValue()) - *((double*)reg2.getValue()));

            if (reg1.type == PT_int)
                port = new int(*((int*)reg1.getValue()) - *((int*)reg2.getValue()));

            task -> setReg(0, port);
        }
        break;

        // rmul
        case 16:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != reg2.type ||
                reg1.type > PT_int)
            {
                errorCommit("Incorrect types!",
                            {"function: rmul",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            port_t port;
            if (reg1.type == PT_float)
                port = new double(*((double*)reg1.getValue()) * *((double*)reg2.getValue()));

            if (reg1.type == PT_int)
                port = new int(*((int*)reg1.getValue()) * *((int*)reg2.getValue()));

            task -> setReg(0, port);
        }
        break;

        // rdiv
        case 17:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != reg2.type ||
                reg1.type > PT_int)
            {
                errorCommit("Incorrect types!",
                            {"function: rdiv",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            port_t port;
            if (reg1.type == PT_float)
                port = new double(*((double*)reg1.getValue()) / *((double*)reg2.getValue()));

            if (reg1.type == PT_int)
                port = new int(*((int*)reg1.getValue()) / *((int*)reg2.getValue()));

            task -> setReg(0, port);
        }
        break;

        //rcce
        case 18:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != reg2.type ||
                reg1.type != PT_ptr)
            {
                errorCommit("Incorrect types!",
                            {"function: rcce",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            port_t port;
            uint16_t    length1 = reg1.length,
                        length2 = reg2.length;
            port.length = length1 + length2;
            uint8_t*    b1 = (uint8_t*) reg1.getValue();
            uint8_t*    b2 = (uint8_t*) reg2.getValue();

            uint8_t*     buf = new uint8_t[port.length];

                // Copy data into new buffer
            for (uint16_t i = 0; i < length1; ++i)
                buf[i] = b1[i];
            for (uint16_t i = 0; i < length2; ++i)
                buf[length1 + i] = b2[i];
            port = buf;

            task -> setReg(0, port);
        }
        break;

        //rccf
        case 19:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != reg2.type ||
                reg1.type != PT_ptr)
            {
                errorCommit("Incorrect types!",
                            {"function: rccf",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            port_t port;
            uint16_t    length1 = reg1.length,
                        length2 = reg2.length;
            port.length = length1 + length2;
            uint8_t*    b1 = (uint8_t*) reg1.getValue();
            uint8_t*    b2 = (uint8_t*) reg2.getValue();

            uint8_t*     buf = new uint8_t[port.length];

                // Copy data into new buffer
            for (uint16_t i = 0; i < length2; ++i)
                buf[i] = b1[i];
            for (uint16_t i = 0; i < length1; ++i)
                buf[length2 + i] = b2[i];
            port = buf;

            task -> setReg(0, port);
        }
        break;

        //rbsl
        case 20:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != reg2.type ||
                reg1.type != PT_int)
            {
                errorCommit("Incorrect types!",
                            {"function: rbsl",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            int value = *((int*)reg1.getValue()) << *((int*)reg2.getValue());
            port_t port;
            port = new int (value);
            task -> setReg(0, port);
        }
        break;

        //rbsr
        case 21:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != reg2.type ||
                reg1.type != PT_int)
            {
                errorCommit("Incorrect types!",
                            {"function: rbsr",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            int value = *((int*)reg1.getValue()) >> *((int*)reg2.getValue());
            port_t port;
            port = new int (value);
            task -> setReg(0, port);
        }
        break;

        //rand
        case 22:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != reg2.type ||
                reg1.type != PT_int)
            {
                errorCommit("Incorrect types!",
                            {"function: rand",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            int value = *((int*)reg1.getValue()) & *((int*)reg2.getValue());
            port_t port;
            port = new int (value);
            task -> setReg(0, port);
        }
        break;

        //ror
        case 23:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != reg2.type ||
                reg1.type != PT_int)
            {
                errorCommit("Incorrect types!",
                            {"function: ror",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            int value = *((int*)reg1.getValue()) | *((int*)reg2.getValue());
            port_t port;
            port = new int (value);
            task -> setReg(0, port);
        }
        break;

        //rxor
        case 24:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != reg2.type ||
                reg1.type != PT_int)
            {
                errorCommit("Incorrect types!",
                            {"function: rxor",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            int value = *((int*)reg1.getValue()) ^ *((int*)reg2.getValue());
            port_t port;
            port = new int (value);
            task -> setReg(0, port);
        }
        break;

        //rfrac
        case 25:
        {
            port_t reg1 = task -> getReg(task -> getBufU8());

            if (reg1.type != PT_float)
            {
                errorCommit("Expected type float!",
                            {"function: rfrac",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            double value = *((double*) reg1.getValue());
            value -= int(value);
            port_t port;
            port = new double (value);
            task -> setReg(0, port);
        }
        break;

        // ritof
        case 26:
        {
            port_t reg1 = task -> getReg(task -> getBufU8());

            if (reg1.type != PT_int)
            {
                errorCommit("Expected type int!",
                            {"function: ritof",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            port_t port;
            port = new double(*((int*) reg1.getValue()));
            task -> setReg(0, port);
        }
        break;

        // rftoi
        case 27:
        {
            port_t reg1 = task -> getReg(task -> getBufU8());

            if (reg1.type != PT_float)
            {
                errorCommit("Expected type float!",
                            {"function: rftoi",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            port_t port;
            port = new int(floor(*((double*) reg1.getValue())) + .01);
            task -> setReg(0, port);
        }
        break;

        //rstorloc
        case 28:
        {
            port_t port;
            port = new int(task -> getSeek());
            task -> setReg(0, port);
        }
        break;

        // rplen
        case 29:
        {
            port_t reg1 = task -> getReg(task -> getBufU8());

            if (reg1.type != PT_ptr)
            {
                errorCommit("Cannot check length of non-pointer value!",
                            {"function: rplen",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            port_t port;
            port = new int(reg1.length);
            task -> setReg(0, port);
        }
        break;

        //rpgb
        case 30:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != PT_ptr)
            {
                errorCommit("Expected type pointer!",
                            {"function: rpgb",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (reg2.type != PT_int)
            {
                errorCommit("Expected type int!",
                            {"function: rpgb",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (*((int*)reg2.getValue()) < 0)
            {
                errorCommit("Register out-of-bounds [negative number]!",
                            {"function: rpgb",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (*((int*)reg2.getValue()) > reg1.length - 1)
            {
                errorCommit((string("Register out-of-bounds!") +
                            "\n\tseekpos: " + std::to_string(task -> getSeek()) +
                            "\n\tbounds: 0, " + std::to_string(reg1.length) + " - 1" +
                            "\n\tlocation: " + std::to_string(*((int*)reg2.getValue()))).c_str());

                stopThread();
                break;
            }

            port_t port;
            port = new int (((uint8_t*) reg1.getValue())[*((int*)reg2.getValue())]);
            task -> setReg(0, port);
        }
        break;

        //rpgi
        case 31:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != PT_ptr)
            {
                errorCommit("Expected type pointer!",
                            {"function: rpsi",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (reg2.type != PT_int)
            {
                errorCommit("Expected type int!",
                            {"function: rpgi",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (*((int*)reg2.getValue()) < 0)
            {
                errorCommit("Register out-of-bounds [negative number]!",
                            {"function: rpgi",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (*((int*)reg2.getValue()) > reg1.length - 4)
            {
                errorCommit((string("Register out-of-bounds!") +
                            "\n\tseekpos: " + std::to_string(task -> getSeek()) +
                            "\n\tbounds: 0, " + std::to_string(reg1.length) + " - 4" +
                            "\n\tlocation: " + std::to_string(*((int*)reg2.getValue()))).c_str());

                stopThread();
                break;
            }

            port_t port;
            port = new int (((int*)(&(((uint8_t*) reg1.getValue())[*((int*)reg2.getValue())])))[0]);
            task -> setReg(0, port);
        }
        break;

        //rpgf
        case 32:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N);

            if (reg1.type != PT_ptr)
            {
                errorCommit("Expected type pointer!",
                            {"function: rpgf",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (reg2.type != PT_int)
            {
                errorCommit("Expected type int!",
                            {"function: rpgf",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (*((int*)reg2.getValue()) < 0)
            {
                errorCommit("Register out-of-bounds [negative number]!",
                            {"function: rpgf",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (*((int*)reg2.getValue()) > reg1.length - 8)
            {
                errorCommit((string("Register out-of-bounds!") +
                            "\n\tseekpos: " + std::to_string(task -> getSeek())+
                            "\n\tbounds: 0, " + std::to_string(reg1.length) + " - 8"+
                            "\n\tlocation: " + std::to_string(*((int*)reg2.getValue()))).c_str());

                stopThread();
                break;
            }

            port_t port;
            port = new double (((double*)(&(((uint8_t*) reg1.getValue())[*((int*)reg2.getValue())])))[0]);
            task -> setReg(0, port);
        }
        break;

        // sqrt
        case 33:
        {
            port_t reg1 = task -> getReg(task -> getBufU8());

            if (reg1.type >= PT_int)
            {
                errorCommit("Incorrect type!",
                            {"function: sqrt",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            port_t port;
            if (reg1.type == PT_float)
                port = new double(sqrt(*((double*) reg1.getValue())));

            if (reg1.type == PT_int)
                port = new int(sqrt(*((int*) reg1.getValue())));

            task -> setReg(0, port);
        }
        break;

        // rpsb
        case 50:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8(),
                    reg3N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N),
                    reg3 = task -> getReg(reg3N);

            if (reg1.type != PT_ptr)
            {
                errorCommit("Expected type pointer!",
                            {"function: rpsb",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (reg2.type != PT_int ||
                reg3.type != PT_int)
            {
                errorCommit("Incorrect type!",
                            {"function: rpsb",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (*((int*)reg2.getValue()) < 0)
            {
                errorCommit("Register out-of-bounds [negative number]!",
                            {"function: rpsb",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (*((int*)reg2.getValue()) > reg1.length - 1)
            {
                errorCommit((string("Register out-of-bounds!")+
                            "\n\tseekpos: " + std::to_string(task -> getSeek())+
                            "\n\tbounds: 0, " + std::to_string(reg1.length) + " - 1"+
                            "\n\tlocation: " + std::to_string(*((int*)reg2.getValue()))).c_str());

                stopThread();
                break;
            }

            uint32_t pos = *((int*)reg2.getValue());
            uint8_t  val = *((int*)reg3.getValue());

            uint8_t* buf = (uint8_t*) reg1.getValue();
            buf[pos] = val;
        }
        break;

         // rpsi
        case 51:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8(),
                    reg3N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N),
                    reg3 = task -> getReg(reg3N);

            if (reg1.type != PT_ptr)
            {
                errorCommit("Expected type pointer!",
                            {"function: rpsi",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (reg2.type != PT_int ||
                reg3.type != PT_int)
            {
                errorCommit("Expected type int!",
                            {"function: rpsi",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (*((int*)reg2.getValue()) < 0)
            {
                errorCommit("Register out-of-bounds [negative number]!",
                            {"function: rpsi",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (*((int*)reg2.getValue()) > reg1.length - 4)
            {
                errorCommit((string("Register out-of-bounds!")+
                            "\n\tseekpos: " + std::to_string(task -> getSeek())+
                            "\n\tbounds: 0, " + std::to_string(reg1.length) + " - 4"+
                            "\n\tlocation: " + std::to_string(*((int*)reg2.getValue()))).c_str());

                stopThread();
                break;
            }

            uint32_t pos = *((int*)reg2.getValue());
            int  val = *((int*)reg3.getValue());

            uint8_t* buf_ = (uint8_t*) reg1.getValue();
            int*     buf = (int*) &(buf_[pos]);
            buf[0] = val;
        }
        break;

         // rpsf
        case 52:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8(),
                    reg3N = task -> getBufU8();

            port_t  reg1 = task -> getReg(reg1N),
                    reg2 = task -> getReg(reg2N),
                    reg3 = task -> getReg(reg3N);

            if (reg1.type != PT_ptr)
            {
                errorCommit("Expected type pointer!");
                stopThread();
                break;
            }

            if (reg2.type != PT_int ||
                reg3.type != PT_float)
            {
                errorCommit("Incorrect type!",
                            {"function: rpsf",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (*((int*)reg2.getValue()) < 0)
            {
                errorCommit("Register out-of-bounds [negative number]!",
                            {"function: rpsf",
                             "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            if (*((int*)reg2.getValue()) > reg1.length - 8)
            {
                errorCommit((string("Register out-of-bounds!")+
                            "\n\tseekpos: " + std::to_string(task -> getSeek())+
                            "\n\tbounds: 0, " + std::to_string(reg1.length) + " - 8"+
                            "\n\tlocation: " + std::to_string(*((int*)reg2.getValue()))).c_str());

                stopThread();
                break;
            }

            uint32_t pos = *((int*)reg2.getValue());
            double  val = *((double*)reg3.getValue());

            uint8_t* buf_ = (uint8_t*) reg1.getValue();
            double*     buf = (double*) &(buf_[pos]);
            buf[0] = val;
        }
        break;

        // rpalloc
        case 53:
        {
            uint8_t reg1N = task -> getBufU8(),
                    reg2N = task -> getBufU8();

            port_t  reg2 = task -> getReg(reg2N);

            if (reg2.type != PT_int)
            {
                errorCommit("Expected type int!",
                            {"function: rpalloc",
                            "seekpos: " + to_string(task -> getSeek())});
                stopThread();
                break;
            }

            port_t port;
            uint8_t* buf = new uint8_t[*((int*)reg2.getValue())];
            for (int i = 0; i < *((int*)reg2.getValue()); ++i)
                buf[i] = 0;

            port = buf;
            port.length = *((int*)reg2.getValue());

            task -> setReg(reg1N, port);
        }
        break;

        // dumplog
        case 100:
        {
            string str = "[BIC] DUMP LOG!\n\tseekpos: " + std::to_string(task -> getSeek()) + "\n\t  REGISTERS";
            for (int i = 0; i < task -> getRegCount(); ++i)
            {
                port_t reg = task -> getReg(i);
                str += "\n\t" + std::to_string(i) + ": ";
                if (reg.type == PT_float)
                    str += "FLOAT, " + std::to_string(*((double*)reg.getValue()));
                else if (reg.type == PT_int)
                    str += "INT, " + std::to_string(*((int*)reg.getValue()));
                else if (reg.type == PT_ptr)
                {
                    str += "PTR, [";
                    for (int i = 0; i < std::min(5, (int) reg.length); ++i)
                    {
                        if (i == 0)
                            str += std::to_string(((uint8_t*) reg.getValue())[i]);
                        else
                            str += ", " + std::to_string(((uint8_t*) reg.getValue())[i]);
                    }
                    if (reg.length > 5)
                        str += ", ...]";
                    else
                        str += "]";
                }
                else
                    str += "UNDF, UNDF";

            }
            str += "\n  IO";
            for (int i = 0; i < task -> getIOCount(); ++i)
            {
                port_t io = task -> getIO(i);
                str += "\n\t" + std::to_string(i) + ": ";
                if (io.type == PT_float)
                    str += "FLOAT, " + std::to_string(*((double*)io.getValue()));
                else if (io.type == PT_int)
                    str += "INT, " + std::to_string(*((int*)io.getValue()));
                else if (io.type == PT_ptr)
                {
                    str += "PTR, [";
                    for (int i = 0; i < std::min(5, (int) io.length); ++i)
                    {
                        if (i == 0)
                            str += std::to_string(((uint8_t*) io.getValue())[i]);
                        else
                            str += ", " + std::to_string(((uint8_t*) io.getValue())[i]);
                    }
                    if (io.length > 5)
                        str += ", ...]";
                    else
                        str += "]";
                }
                else
                    str += "UNDF, UNDF";

            }

            messageCommit(str.c_str());
        }
        break;

        // Trace
        case 101:
        {
            uint8_t reg1N = task -> getBufU8(),
                    traceTY = task -> getBufU8(),
                    traceTG = task -> getBufU8();

            //messageCommit(("[BIC] Trace toggle @ " + std::to_string(task -> getSeek())).c_str());

            if (traceTY == 0)
            {
                if (traceTG == 1)
                    task -> addTraceReg(reg1N);
                else
                    task -> pullTraceReg(reg1N);
            }
            else
            {
                if (traceTG == 1)
                    task -> addTraceIO(reg1N);
                else
                    task -> pullTraceIO(reg1N);
            }
        }
        break;

        default:
            string str = "Unknown command [" + std::to_string(command) + "]";
            str += "\n\tseekpos: " + std::to_string(task -> getSeek());
            errorCommit(str.c_str());
        break;
    }
}

void BIC_Runner::threadExecute(BIC_Runner* runner)
{
    messageCommit("[BIC] Thread successfully created!");
    BIC_Container* task;
    //while (runner -> getTaskCount() > 0 && runner -> getRunning())
    for (uint32_t i = 0; runner -> getTaskCount() > 0 && runner -> getRunning();
         i = (i + 1 >= runner -> getTaskCount()) ? 0 : i + 1)
    {
        //task = runner -> getCurrentTask();
        task = runner -> getTask(i);
        /*if (task == nullptr)
        {
            errorCommit("Unexpected error pulling task! Terminating thread.");
            break;
        }*/

        runner -> taskCalculateBit(task);
    }

    messageCommit("[BIC] Thread terminated...");
    runner -> stopThread();
}
